
# This file was generated by 'versioneer.py' (0.21) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2021-10-25T11:01:43+0200",
 "dirty": false,
 "error": null,
 "full-revisionid": "c3b3749c1fd9e5171da9313a9a630a3b54812700",
 "version": "0.30.0.dev0"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
