body{font-family:Source Sans Pro,Trebuchet MS,Lucida Grande,Bitstream Vera Sans,Helvetica Neue,sans-serif;font-size:13px;margin:0;text-align:center;color:#293c4b}.header{display:-webkit-box;display:-webkit-flex;display:-ms-flexbox;display:flex;-webkit-flex-wrap:wrap;-ms-flex-wrap:wrap;flex-wrap:wrap}.parameters{margin:1em}.table-fill{background:#fff;border-radius:8px;border-collapse:collapse;height:320px;margin:auto;max-width:600px;padding:5px;width:100%;box-shadow:0 5px 10px rgba(0,0,0,.1)}.parameters th{color:#d5dde5;background:#1b1e24;border-bottom:4px solid #9ea7af;border-right:1px solid #343a45;font-size:13px;font-weight:100;padding:5px;text-align:left;text-shadow:0 1px 1px rgba(0,0,0,.1);vertical-align:middle}.parameters th:first-child{border-top-left-radius:3px}.parameters th:last-child{border-top-right-radius:3px;border-right:none}.parameters tr{border-top:1px solid #c1c3d1;border-bottom-:1px solid #c1c3d1;color:#666b85;font-size:13px;font-weight:400;text-shadow:0 1px 1px hsla(0,0%,100%,.1)}.parameters tr:hover td{background:#4e5066;color:#fff;border-top:1px solid #22262e;border-bottom:1px solid #22262e}.parameters tr:first-child{border-top:none}.parameters tr:last-child{border-bottom:none}.parameters tr:nth-child(odd) td{background:#ebebeb}.parameters tr:nth-child(odd):hover td{background:#4e5066}.parameters tr:last-child td:first-child{border-bottom-left-radius:3px}.parameters tr:last-child td:last-child{border-bottom-right-radius:3px}.parameters td{background:#fff;padding:3px;text-align:left;vertical-align:middle;font-weight:300;font-size:13px;text-shadow:-1px -1px 1px rgba(0,0,0,.1);border-right:1px solid #c1c3d1}.parameters td:last-child{border-right:0}.instrument{margin:5px;text-align:left}.instrument :hover{background:#4e5066;color:#fff}.parametersContainer{display:-webkit-box;display:-webkit-flex;display:-ms-flexbox;display:flex;-webkit-flex-wrap:wrap;-ms-flex-wrap:wrap;flex-wrap:wrap}
