from qcodes.instrument.delegate.delegate_instrument import (
    GroupedParameter,
    DelegateInstrument
)
from qcodes.instrument.delegate.instrument_group import InstrumentGroup

from qcodes.instrument.delegate.delegate_channel_instrument import (
    DelegateChannelInstrument
)
